# (c) 2015, Ian Clegg <ian.clegg@sourcewarp.com>
#
# winrmlib is licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
__author__ = 'ian.clegg@sourcewarp.com'

class ResourceLocator(object):
    """
    Resource Locator
    TO DO, constructor should accept a dictionary of options and a dictionary of list for selectors
    """

    def __init__(self, url):
        self.url = url
        self.options = {}
        self.selectors = {}

    def add_option(self, name, value, must_comply):
        self.options[name] = [value, must_comply]

    def clear_options(self):
        self.options = {}

    def add_selector(self, name, value):
        self.selectors[name] = value

    def clear_selectors(self):
        self.selectors = {}


# (c) 2015, Ian Clegg <ian.clegg@sourcewarp.com>
#
# winrmlib is licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
__author__ = 'ian.clegg@sourcewarp.com'

# The source code in this file was inspired by the work of Matthieu Suiche,
# http://sandman.msuiche.net/, and the information presented released as
# part of the Microsoft Interoperability Initiative:
# http://download.microsoft.com/download/a/e/6/ae6e4142-aa58-45c6-8dcf-a657e5900cd3/%5BMS-DRSR%5D.pdf

from struct import unpack
from struct import error as StructError

def recombine(outbuf):
    return "".join(outbuf[k] for k in sorted(outbuf.keys()))

def xpress_decode(input_buffer):
    output_buffer_ = {}
    output_index = 0
    input_index = 0
    indicator_bit = 0
    nibble_index = 0

    # we are decoding the entire input here, so I have changed
    # the check to see if we're at the end of the output buffer
    # with a check to see if we still have any input left.
    while input_index < len(input_buffer):
        if (indicator_bit == 0):
            # in pseudocode this was indicator_bit = ..., but that makes no
            # sense, so I think this was intended...
            try:
                indicator = unpack("<L", input_buffer[input_index:input_index + 4])[0]
            except StructError:
                return recombine(output_buffer_)

            input_index += 4
            indicator_bit = 32

        indicator_bit = indicator_bit - 1
        # check whether the bit specified by indicator_bit is set or not
        # set in indicator. For example, if indicator_bit has value 4
        # check whether the 4th bit of the value in indicator is set
        if not (indicator & (1 << indicator_bit)):
            try:
                output_buffer_[output_index] = input_buffer[input_index]
            except IndexError:
                return recombine(output_buffer_)

            input_index += 1
            output_index += 1
        else:
            # Get the length. This appears to use a scheme whereby if
            # the value at the current width is all ones, then we assume
            # that it is actually wider. First we try 3 bits, then 3
            # bits plus a nibble, then a byte, and finally two bytes (an
            # unsigned short). Also, if we are using a nibble, then every
            # other time we get the nibble from the high part of the previous
            # byte used as a length nibble.
            # Thus if a nibble byte is F2, we would first use the low part (2),
            # and then at some later point get the nibble from the high part (F).

            try:
                length = unpack("<H", input_buffer[input_index:input_index + 2])[0]
            except StructError:
                return recombine(output_buffer_)

            input_index += 2
            offset = length / 8
            length = length % 8
            if length == 7:
                if nibble_index == 0:
                    nibble_index = input_index
                    length = ord(input_buffer[input_index]) % 16
                    input_index += 1
                else:
                    # get the high nibble of the last place a nibble sized
                    # length was used thus we don't waste that extra half
                    # byte :p
                    length = ord(input_buffer[nibble_index]) / 16
                    nibble_index = 0

                if length == 15:
                    length = ord(input_buffer[input_index])
                    input_index += 1
                    if length == 255:
                        try:
                            length = unpack("<H", input_buffer[input_index:input_index + 2])[0]
                        except StructError:
                            return recombine(output_buffer_)
                        input_index = input_index + 2
                        length = length - (15 + 7)
                    length = length + 15
                length = length + 7
            length = length + 3

            while length != 0:
                try:
                    output_buffer_[output_index] = output_buffer_[output_index - offset - 1]
                except KeyError:
                    return recombine(output_buffer_)
                output_index += 1
                length -= 1

    return recombine(output_buffer_)


class Compressor:

	def __init__(self):
		self.referencePrefix = "`"
		self.referencePrefixCode = ord(self.referencePrefix)
		self.referenceIntBase = 96
		self.referenceIntFloorCode = ord(" ")
		self.referenceIntCeilCode = self.referenceIntFloorCode + self.referenceIntBase - 1
		self.maxStringDistance = self.referenceIntBase ** 2 - 1
		self.minStringLength = 5
		self.maxStringLength = self.referenceIntBase ** 1 - 1 + self.minStringLength
		self.maxWindowLength = self.maxStringDistance + self.minStringLength;
		self.defaultWindowLength = 144

	def compress(self, data, windowLength = None):
		"""Compresses text data using the LZ77 algorithm."""

		if windowLength == None:
			windowLength = self.defaultWindowLength

		compressed = ""
		pos = 0
		lastPos = len(data) - self.minStringLength

		while pos < lastPos:

			searchStart = max(pos - windowLength, 0);
			matchLength = self.minStringLength
			foundMatch = False
			bestMatchDistance = self.maxStringDistance
			bestMatchLength = 0
			newCompressed = None

			while (searchStart + matchLength) < pos:

				m1 = data[searchStart : searchStart + matchLength]
				m2 = data[pos : pos + matchLength]
				isValidMatch = (m1 == m2 and matchLength < self.maxStringLength)

				if isValidMatch:
					matchLength += 1
					foundMatch = True
				else:
					realMatchLength = matchLength - 1

					if foundMatch and realMatchLength > bestMatchLength:
						bestMatchDistance = pos - searchStart - realMatchLength
						bestMatchLength = realMatchLength

					matchLength = self.minStringLength
					searchStart += 1
					foundMatch = False

			if bestMatchLength:
				newCompressed = (self.referencePrefix + self.__encodeReferenceInt(bestMatchDistance, 2) + self.__encodeReferenceLength(bestMatchLength))
				pos += bestMatchLength
			else:
				if data[pos] != self.referencePrefix:
					newCompressed = data[pos]
				else:
					newCompressed = self.referencePrefix + self.referencePrefix
				pos += 1

			compressed += newCompressed

		return compressed + data[pos:].replace("`", "``")

	def decompress(self, data):
		"""Decompresses LZ77 compressed text data"""

		decompressed = ""
		pos = 0
		while pos < len(data):
			currentChar = data[pos]
			if currentChar != self.referencePrefix:
				decompressed += currentChar
				pos += 1
			else:
				nextChar = data[pos + 1]
				if nextChar != self.referencePrefix:
					distance = self.__decodeReferenceInt(data[pos + 1 : pos + 3], 2)
					length = self.__decodeReferenceLength(data[pos + 3])
					start = len(decompressed) - distance - length
					end = start + length
					decompressed += decompressed[start : end]
					pos += self.minStringLength - 1
				else:
					decompressed += self.referencePrefix
					pos += 2

		return decompressed

	def __encodeReferenceInt(self, value, width):
		if value >= 0 and value < (self.referenceIntBase ** width - 1):
			encoded = ""
			while value > 0:
				encoded = chr((value % self.referenceIntBase) + self.referenceIntFloorCode) + encoded
				value = int(value / self.referenceIntBase)

			missingLength = width - len(encoded)
			for i in range(missingLength):
				encoded = chr(self.referenceIntFloorCode) + encoded

			return encoded
 		else:
			raise Exception("Reference value out of range: %d (width = %d)" % (value, width))

	def __encodeReferenceLength(self, length):
		return self.__encodeReferenceInt(length - self.minStringLength, 1)

	def __decodeReferenceInt(self, data, width):
		value = 0
		for i in range(width):
			value *= self.referenceIntBase
			charCode = ord(data[i])
			if charCode >= self.referenceIntFloorCode and charCode <= self.referenceIntCeilCode:
				value += charCode - self.referenceIntFloorCode
			else:
				raise Exception("Invalid char code: %d" % charCode)

		return value

	def __decodeReferenceLength(self, data):
		return self.__decodeReferenceInt(data, 1) + self.minStringLength


# (c) 2015, Ian Clegg <ian.clegg@sourcewarp.com>
#
# winrmlib is licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
__author__ = 'ian.clegg@sourcewarp.com'

import base64
import logging

from collections import OrderedDict
from api.resourcelocator import ResourceLocator
from api.session import Session

# compression suport still in devel
# import api.compression


class CommandShell(object):

    def __init__(self, endpoint, username, password, **kwargs):
        """
        timeout
        codepage
        noprofile
        environment
        """
        # Process arguments
        self.environment = kwargs.get('environment', None)
        self.working_directory = kwargs.get('working_directory', None)
        self.idle_timeout = kwargs.get('idle_timeout', 180000)
        codepage = kwargs.get('codepage', 437)

        # Build the Session and the SOAP Headers
        self.shell_id = None
        self.session = Session(endpoint, 'ntlm', username, password)
        self.resource = ResourceLocator(CommandShell.ShellResource)
        self.resource.add_option('WINRS_CODEPAGE', codepage, True)
        if bool(kwargs.get('noprofile', False)):
            self.resource.add_option('WINRS_NOPROFILE', 'FALSE', True)
        else:
            self.resource.add_option('WINRS_NOPROFILE', 'TRUE', True)

    def open(self, input_streams=['stdin'], output_streams=['stderr', 'stdout']):
        """
        Opens the remote shell
        """
        shell = dict()
        shell['rsp:InputStreams'] = " ".join(input_streams)
        shell['rsp:OutputStreams'] = " ".join(output_streams)
        shell['rsp:IdleTimeout'] = str(self.idle_timeout)

        if self.working_directory is not None:
            shell['rsp:WorkingDirectory'] = str(self.working_directory)

        if self.environment is not None:
            variables = []
            for key, value in self.environment.items():
                variables.append({'#text': str(value), '@Name': key})
            shell['rsp:Environment'] = {'Variable': variables}

        response = self.session.create(self.resource, {'rsp:Shell': shell})
        self.shell_id = response['rsp:Shell']['rsp:ShellId']

    def run(self, command, arguments=(), console_mode_stdin=True, skip_cmd_shell=False):
        """This function does something.
        :param command: The command to be executed
        :type name: str.
        :param arguments: A list of arguments to be passed to the command
        :type state: str.
        :returns:  int -- the return code.
        :raises: AttributeError, KeyError

        iclegg: blocking i/o operations are slow, doesnt Python have a moden 'async' mechanism
        rather than replying on 80's style callbacks?
        """
        logging.info('running command: ' + command)
        resource = ResourceLocator(CommandShell.ShellResource)
        resource.add_selector('ShellId', self.shell_id)
        resource.add_option('WINRS_SKIP_CMD_SHELL', ['FALSE', 'TRUE'][bool(skip_cmd_shell)], True)
        resource.add_option('WINRS_CONSOLEMODE_STDIN', ['FALSE', 'TRUE'][bool(console_mode_stdin)], True)

        command = OrderedDict([('rsp:Command', command)])
        for argument in arguments:
            command['rsp:Arguments'] = argument

        response = self.session.command(resource, {'rsp:CommandLine': command})
        command_id = response['rsp:CommandResponse']['rsp:CommandId']
        logging.info('receive command: ' + command_id)
        return command_id

    def receive(self, command_id, streams=('stdout', 'stderr'), command_timeout=60):
        """
        Recieves data
        :param command_id:
        :param streams:
        :param command_timeout:
        :return:
        """
        logging.info('receive command: ' + command_id)
        (session_streams, exit_code, done) = self._receive_once(command_id, streams)
        complete_streams = session_streams
        while not done:
            (session_streams, exit_code, done) = self._receive_once(command_id, streams)
            for stream_name in session_streams:
                complete_streams[stream_name] += session_streams[stream_name]

        if sorted(complete_streams.keys()) == sorted(['stderr', 'stdout']):
            return complete_streams['stdout'], complete_streams['stderr'], exit_code
        else:
            return complete_streams, exit_code

    def _receive_once(self, command_id, streams=('stdout', 'stderr')):
        """
        Recieves data
        :param command_id:
        :param streams:
        :return:
        """
        logging.info('receive command: ' + command_id)
        resource = ResourceLocator(CommandShell.ShellResource)
        resource.add_selector('ShellId', self.shell_id)

        stream_attributes = {'#text': " ".join(streams), '@CommandId': command_id}
        receive = {'rsp:Receive': {'rsp:DesiredStream': stream_attributes}}
        response = self.session.recieve(resource, receive)['rsp:ReceiveResponse']

        decoded_streams = {}
        for stream in streams:
            decoded_streams[stream] = ''

        stream = response['rsp:Stream']
        if isinstance(stream, list):
            response_streams = stream
        else:
            response_streams = [response.Stream]

        for stream in response_streams:
            if stream['@CommandId'] == command_id and '#text' in stream:
                decoded_streams[stream['@Name']] += base64.b64decode(stream['#text'])
                # XPRESS Compression Testing
                # print "\\x".join("{:02x}".format(ord(c)) for c in base64.b64decode(stream['#text']))
                # data = base64.b64decode(stream['#text'])
                # f = open('c:\\users\\developer\\temp\\data.bin', 'wb')
                # f.write(data)
                # f.close()
                # decode = api.compression.xpress_decode(data[4:])
        exit_code = None
        done = response['rsp:CommandState']['@State'] == CommandShell.StateDone
        if done:
            exit_code = int(response['rsp:CommandState']['rsp:ExitCode'])

        return decoded_streams, exit_code, done

    def close(self):
        """
        Closes pipe
        :return:
        """
        resource = ResourceLocator(CommandShell.ShellResource)
        resource.add_selector('ShellId', self.shell_id)
        self.session.delete(resource)


# Namespaces
CommandShell.ShellResource = 'http://schemas.microsoft.com/wbem/wsman/1/windows/shell/cmd'

# Command States
CommandShell.StateDone = 'http://schemas.microsoft.com/wbem/wsman/1/windows/shell/CommandState/Done'
CommandShell.StatePending = 'http://schemas.microsoft.com/wbem/wsman/1/windows/shell/CommandState/Pending'
CommandShell.StateRunning = 'http://schemas.microsoft.com/wbem/wsman/1/windows/shell/CommandState/Running'


import os
import sys
import urllib
from setuptools import setup
from distutils.core import Command

__version__ = '0.0.3'
project_name = 'winrmlib'

# PyPi supports only reStructuredText, so pandoc should be installed
# before uploading package
try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except ImportError:
    long_description = ''


class BootstrapEnvironmentCommand(Command):
    description = 'create project development environment from scratch'
    user_options = []

    def __init__(self, dist):
        Command.__init__(self, dist)
        self.project_dir = os.path.abspath(os.path.dirname(__file__))
        self.temp_dir = os.path.join(self.project_dir, '~temp')
        self.virtualenv_dir = os.path.join(self.project_dir, 'env')

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        self.setup_virtual_environment()

    def setup_virtual_environment(self):
        """Creates Python env if it not exists or re-create if outdated."""
        if not os.path.exists(self.temp_dir):
            os.makedirs(self.temp_dir)

        if not os.path.exists(self.virtualenv_dir):
            virtualenv_setup_path = os.path.join(self.temp_dir, 'virtualenv.py')
            urllib.urlretrieve(
                'https://raw.github.com/pypa/virtualenv/master/virtualenv.py',
                virtualenv_setup_path)
            virtualenv_prompt = '\({0}\)'.format(project_name) \
                if os.name == 'posix' else '({0})'.format(project_name)
            os.system('python {0} --prompt={1} env'.format(
                virtualenv_setup_path, virtualenv_prompt))
            self.activate_virtual_environment()
            easy_install_setup_path = os.path.join(self.temp_dir, 'ez_setup.py')
            urllib.urlretrieve(
                'http://peak.telecommunity.com/dist/ez_setup.py',
                easy_install_setup_path)
            os.system(
                'python {0} -U setuptools'.format(easy_install_setup_path))
            os.system(
                'pip install -r requirements.txt --download-cache={0}'.format(
                    self.temp_dir))

    def activate_virtual_environment(self):
        """ Activates virtual environment in the same way as activate.bat or activate.sh from virtualenv package """
        if not hasattr(sys, 'real_prefix') or 'VIRTUAL_ENV' not in os.environ:
            activate_this = '{0}/env/bin/activate_this.py'.format(
                self.project_dir)
            virtual_env_path = '{0}/env/bin:'.format(self.project_dir)
            if os.name != 'posix':
                activate_this = '{0}/env/Scripts/activate_this.py'.format(
                    self.project_dir)
                virtual_env_path = '{0}\env\Scripts;'.format(self.project_dir)

            execfile(activate_this, dict(__file__=activate_this))
            path = os.environ['PATH']
            os.environ['PATH'] = virtual_env_path + path


setup(
    name=project_name,
    version=__version__,
    description='Python library for Windows Remote Management',
    long_description=long_description,
    keywords='winrm winrw ws-man ws-management powershell'.split(' '),
    author='Ian Clegg',
    author_email='ian.clegg@sourcewarp.com',
    url='http://github.com/iclegg/winrmlib/',
    license='MIT license',
    packages=['winrmlib', 'winrmlib.api'],
    package_data={'winrmlib.api': ['assets/xsd/*.xsd', 'assets/*.wsdl']},
    install_requires=[],
    cmdclass={'bootstrap_env': BootstrapEnvironmentCommand},
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Clustering',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Systems Administration',
	],
)
